#################################################
#################################################
##          COPYRIGHT - DATA SOCIETY           ##
#################################################
#################################################

###################################
##    LOGISTIC REGRESSION        ##
###################################

## NOTE: To run code blocks (separated by #%%), select somewhere inside the block 
##       and press ctrl + enter for Windows or Linux, 
##       and command + enter for Macs

#################################################
#%%
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

#%%

# https://rdrr.io/cran/ISLR/man/Default.html
credit_data = pd.read_csv("default.csv")
credit_data.head()

#################################################
#%%
# Slide 20: How to make the scatter plot shown

# source: https://www.kaggle.com/uciml/default-of-credit-card-clients-dataset/data

defaulted = credit_data['default'] == 1
plt.scatter(credit_data.loc[~defaulted]['balance'], 
            credit_data.loc[~defaulted]['income'], 
             c='blue', s=1)
plt.scatter(credit_data.loc[defaulted]['balance'], 
            credit_data.loc[defaulted]['income'], 
             c='red', s=10, edgecolor='white')
plt.xlabel('Balance')
plt.ylabel('Income')

#%%
plt.savefig('plots/default.png', dpi=200)
plt.close()

#################################################
#%%
# Slide 22: Fit a logistic regression using just the balance variable

from sklearn.linear_model import LogisticRegression

model = LogisticRegression() # make a LR model using the default settings
lr_fit = model.fit(X = credit_data[['balance']],
                   y = np.ravel(credit_data['default'])) 

#%%
# see the coefficients (beta's):
beta0 = lr_fit.intercept_[0] # y intercept
beta1 = lr_fit.coef_[0][0] # coefficient of balance
np.exp(beta0)
np.exp(beta1)
#%%
# Slide 23: predict and calculate accuracy 
y_pred = lr_fit.predict(credit_data['balance'].values.reshape(-1,1))

#%%

# correct classification rate (accuracy):
y_data = credit_data['default']
N = len(credit_data)
sum(y_data==y_pred)/N

#################################################
#%%
# Slide 31: what if we just guessed "no default" for everyone?

baserate = sum(y_data==0)/N
print('The base rate of no-default is {}'.format(baserate))
# so by using balance, we do a little better than the "dummy" model 
# that just guesses no

#################################################
#%%
# Slide 35: Split data into test and train

from sklearn.model_selection import train_test_split
X = credit_data[['balance', 'student']]
y_data = credit_data['default']
X_train, X_test, y_train, y_test = train_test_split(X, y_data, 
                                                    test_size=0.2, 
                                                    random_state=0)

lr_fit2_train = LogisticRegression().fit(X_train, y_train)
lr_fit2_train.intercept_
lr_fit2_train.coef_

#################################################
#%%
# Slide 36: how does it do on train and test data?
y_pred2_train = lr_fit2_train.predict(X_train)

sum(y_pred2_train==y_train)/len(y_train)

#%%
# accuracy on test data
y_pred2_test = lr_fit2_train.predict(X_test)
sum(y_pred2_test==y_test)/len(y_test)

#################################################
#%%
# Slide 40: Calculate the ROC Curve
def logistic(a):
    return 1/(1 + np.exp(-a))


# try different thresholds (on test data!)
beta0 = lr_fit2_train.intercept_[0]
beta1 = lr_fit2_train.coef_[0][0]
beta2 = lr_fit2_train.coef_[0][1]
def get_roc(cutoff):
    my_y_pred = []
    n = len(y_test)
    for i in range(n):
        x = X_test.iloc[i][['balance','student']]
        a = beta0 + beta1*x[0] + beta2*x[1]
        p = logistic(a)
        if p < cutoff:
            my_y_pred.append(0)
        else:
            my_y_pred.append(1)
            
    my_y_pred = np.array(my_y_pred)
    tpr = sum((my_y_pred == 1) & (y_test.values == 1))/n
    fpr = sum((my_y_pred == 1) & (y_test.values == 0))/n
    return (tpr, fpr)

#%%
TPR = []
FPR = []
for h in np.linspace(0,1,15):
    print('Testing cutoff = {}'.format(round(h,4)))
    tpr, fpr = get_roc(h)
    TPR.append(tpr)
    FPR.append(fpr)
#%%
plt.plot(FPR, TPR, marker='o')
plt.xlabel('False positive rate')
plt.ylabel('True positive rate')
plt.title('ROC curve: Default ~ Balance + Student')

#%%
# as we let the threshold go lower, we get a perfect
# true positive rate
y_test.mean() # 0.37, which is the max of the ROC plot, corresponding
# to cutoff = 0

#%%
plt.savefig('plots/roc.png', dpi=200)
plt.close()

#################################################
#%%
# Slide 44: Multinomial regression
# read in data
prog_data = pd.read_csv("hsbdemo.csv")
prog_data['female'] = pd.get_dummies(prog_data['gender'])['female']
prog_data = pd.concat([prog_data, pd.get_dummies(prog_data['ses'])], axis = 1)

#################################################
#%%
# Slide 45: Fit the model
# dummy code ses = socio-economic-status
# https://stats.idre.ucla.edu/sas/dae/multinomiallogistic-regression/
X = prog_data[['write', 'math', 'high', 'low', 'middle', 'female']]
y = np.ravel(prog_data['prog'])
multinom = LogisticRegression()
multinom_fit = multinom.fit(X, y)
multinom_fit.intercept_
multinom_fit.coef_
y_pred = multinom_fit.predict(X)

sum(y_pred == y)/len(prog_data)
